import fs from 'fs';
import es from 'event-stream';
import glob from 'glob';
import cliProgress from 'cli-progress';
import elasticsearch from '@elastic/elasticsearch';

var DEFAULT_BUFFER_SIZE = 1000;

function createMappingFactory(ref) {
  var sourceClient = ref.sourceClient;
  var sourceIndexName = ref.sourceIndexName;
  var targetClient = ref.targetClient;
  var targetIndexName = ref.targetIndexName;
  var mappings = ref.mappings;
  var mappingsOverride = ref.mappingsOverride;
  var indexMappingTotalFieldsLimit = ref.indexMappingTotalFieldsLimit;
  var verbose = ref.verbose;

  return async function () {
    var targetMappings = mappingsOverride ? undefined : mappings;

    if (sourceClient && sourceIndexName && typeof targetMappings === 'undefined') {
      try {
        var mapping = await sourceClient.indices.getMapping({
          index: sourceIndexName,
        });
        targetMappings = mapping[sourceIndexName].mappings;
      } catch (err) {
        console.log('Error reading source mapping', err);
        return;
      }
    }

    if (typeof targetMappings === 'object' && targetMappings !== null) {
      if (mappingsOverride) {
        targetMappings = Object.assign({}, targetMappings,
          {properties: Object.assign({}, targetMappings.properties,
            mappings)});
      }

      try {
        var resp = await targetClient.indices.create({
          index: targetIndexName,
          body: Object.assign({}, {mappings: targetMappings},
            (indexMappingTotalFieldsLimit !== undefined
              ? {
                  settings: {
                    'index.mapping.total_fields.limit': indexMappingTotalFieldsLimit,
                  },
                }
              : {})),
        });
        if (verbose) { console.log('Created target mapping', resp); }
      } catch (err) {
        console.log('Error creating target mapping', err);
      }
    }
  };
}

var MAX_QUEUE_SIZE = 15;

function fileReaderFactory(indexer, fileName, transform, splitRegex, verbose) {
  function startIndex(files) {
    var ingestQueueSize = 0;
    var finished = false;

    var file = files.shift();
    var s = fs
      .createReadStream(file)
      .pipe(es.split(splitRegex))
      .pipe(
        es
          .mapSync(function (line) {
            try {
              // skip empty lines
              if (line === '') {
                return;
              }

              var doc =
                typeof transform === 'function'
                  ? JSON.stringify(transform(JSON.parse(line)))
                  : line;

              // if doc is undefined we'll skip indexing it
              if (typeof doc === 'undefined') {
                s.resume();
                return;
              }

              // the transform callback may return an array of docs so we can emit
              // multiple docs from a single line
              if (Array.isArray(doc)) {
                doc.forEach(function (d) { return indexer.add(d); });
                return;
              }

              indexer.add(doc);
            } catch (e) {
              console.log('error', e);
            }
          })
          .on('error', function (err) {
            console.log('Error while reading file.', err);
          })
          .on('end', function () {
            if (verbose) { console.log('Read entire file: ', file); }
            if (files.length > 0) {
              startIndex(files);
              return;
            }

            indexer.finish();
            finished = true;
          })
      );

    indexer.queueEmitter.on('queue-size', async function (size) {
      if (finished) { return; }
      ingestQueueSize = size;

      if (ingestQueueSize < MAX_QUEUE_SIZE) {
        s.resume();
      } else {
        s.pause();
      }
    });

    indexer.queueEmitter.on('resume', function () {
      if (finished) { return; }
      ingestQueueSize = 0;
      s.resume();
    });
  }

  return function () {
    glob(fileName, function (er, files) {
      startIndex(files);
    });
  };
}

var EventEmitter = require('events');

var queueEmitter = new EventEmitter();

var parallelCalls = 1;

// a simple helper queue to bulk index documents
function indexQueueFactory(ref) {
  var client = ref.targetClient;
  var targetIndexName = ref.targetIndexName;
  var bufferSize = ref.bufferSize; if ( bufferSize === void 0 ) bufferSize = DEFAULT_BUFFER_SIZE;
  var skipHeader = ref.skipHeader; if ( skipHeader === void 0 ) skipHeader = false;
  var verbose = ref.verbose; if ( verbose === void 0 ) verbose = true;

  var buffer = [];
  var queue = [];
  var ingesting = 0;
  var ingestTimes = [];
  var finished = false;

  var ingest = function (b) {
    if (typeof b !== 'undefined') {
      queue.push(b);
      queueEmitter.emit('queue-size', queue.length);
    }

    if (ingestTimes.length > 5) { ingestTimes = ingestTimes.slice(-5); }

    if (ingesting < parallelCalls) {
      var docs = queue.shift();

      queueEmitter.emit('queue-size', queue.length);
      if (queue.length <= 5) {
        queueEmitter.emit('resume');
      }

      ingesting += 1;

      if (verbose)
        { console.log(("bulk ingest docs: " + (docs.length / 2) + ", queue length: " + (queue.length))); }

      var start = Date.now();
      client
        .bulk({ body: docs })
        .then(function () {
          var end = Date.now();
          var delta = end - start;
          ingestTimes.push(delta);
          ingesting -= 1;

          var ingestTimesMovingAverage =
            ingestTimes.length > 0
              ? ingestTimes.reduce(function (p, c) { return p + c; }, 0) / ingestTimes.length
              : 0;
          var ingestTimesMovingAverageSeconds = Math.floor(ingestTimesMovingAverage / 1000);

          if (
            ingestTimes.length > 0 &&
            ingestTimesMovingAverageSeconds < 30 &&
            parallelCalls < 10
          ) {
            parallelCalls += 1;
          } else if (
            ingestTimes.length > 0 &&
            ingestTimesMovingAverageSeconds >= 30 &&
            parallelCalls > 1
          ) {
            parallelCalls -= 1;
          }

          if (queue.length > 0) {
            ingest();
          } else if (queue.length === 0 && finished) {
            queueEmitter.emit('finish');
          }
        })
        .catch(function (error) {
          console.error(error);
          ingesting -= 1;
          parallelCalls = 1;
          if (queue.length > 0) {
            ingest();
          }
        });
    }
  };

  return {
    add: function (doc) {
      if (finished) {
        throw new Error('Unexpected doc added after indexer should finish.');
      }

      if (!skipHeader) {
        var header = { index: { _index: targetIndexName } };
        buffer.push(header);
      }
      buffer.push(doc);

      if (queue.length === 0) {
        queueEmitter.emit('resume');
      }

      if (buffer.length >= bufferSize * 2) {
        ingest(buffer);
        buffer = [];
      }
    },
    finish: function () {
      finished = true;

      if (buffer.length > 0) {
        ingest(buffer);
        buffer = [];
      } else if (queue.length === 0 && ingesting === 0) {
        queueEmitter.emit('finish');
      }
    },
    queueEmitter: queueEmitter,
  };
}

var MAX_QUEUE_SIZE$1 = 15;

// create a new progress bar instance and use shades_classic theme
var progressBar = new cliProgress.SingleBar({}, cliProgress.Presets.shades_classic);

function indexReaderFactory(
  indexer,
  sourceIndexName,
  transform,
  client,
  query,
  bufferSize,
  populatedFields
) {
  if ( bufferSize === void 0 ) bufferSize = DEFAULT_BUFFER_SIZE;
  if ( populatedFields === void 0 ) populatedFields = false;

  return async function indexReader() {
    var responseQueue = [];
    var docsNum = 0;

    async function fetchPopulatedFields() {
      try {
        var response = await client.search({
          index: sourceIndexName,
          size: bufferSize,
          query: {
            function_score: {
              query: query,
              random_score: {},
            },
          },
        });

        // Get all field names for each returned doc and flatten it
        // to a list of unique field names used across all docs.
        return new Set(response.hits.hits.map(function (d) { return Object.keys(d._source); }).flat(1));
      } catch (e) {
        console.log('error', e);
      }
    }

    function search(fields) {
      return client.search(Object.assign({}, {index: sourceIndexName,
        scroll: '600s',
        size: bufferSize,
        query: query},
        (fields ? { _source: fields } : {})));
    }

    function scroll(id) {
      return client.scroll({
        scroll_id: id,
        scroll: '600s',
      });
    }

    var fieldsWithData;

    // identify populated fields
    if (populatedFields) {
      fieldsWithData = await fetchPopulatedFields();
      console.log('fieldsWithData', fieldsWithData);
    }

    // start things off by searching, setting a scroll timeout, and pushing
    // our first response into the queue to be processed
    var se = await search(fieldsWithData);
    responseQueue.push(se);
    progressBar.start(se.hits.total.value, 0);
    console.log('se', se.hits.hits[0]);

    function processHit(hit) {
      docsNum += 1;
      try {
        var doc = typeof transform === 'function' ? transform(hit._source) : hit._source; // eslint-disable-line no-underscore-dangle
        // console.log('doc', doc);

        // if doc is undefined we'll skip indexing it
        if (typeof doc === 'undefined') {
          return;
        }

        // the transform callback may return an array of docs so we can emit
        // multiple docs from a single line
        if (Array.isArray(doc)) {
          doc.forEach(function (d) { return indexer.add(d); });
          return;
        }

        indexer.add(doc);
      } catch (e) {
        console.log('error', e);
      }
    }

    var ingestQueueSize = 0;
    var scrollId = se._scroll_id; // eslint-disable-line no-underscore-dangle
    var readActive = false;

    async function processResponseQueue() {
      while (responseQueue.length) {
        readActive = true;
        var response = responseQueue.shift();

        // collect the docs from this response
        response.hits.hits.forEach(processHit);

        progressBar.update(docsNum);

        // check to see if we have collected all of the docs
        if (response.hits.total.value === docsNum) {
          indexer.finish();
          break;
        }

        if (ingestQueueSize < MAX_QUEUE_SIZE$1) {
          // get the next response if there are more docs to fetch
          var sc = await scroll(response._scroll_id); // eslint-disable-line no-await-in-loop,no-underscore-dangle,max-len
          scrollId = sc._scroll_id; // eslint-disable-line no-underscore-dangle
          responseQueue.push(sc);
        } else {
          readActive = false;
        }
      }
    }

    indexer.queueEmitter.on('queue-size', async function (size) {
      ingestQueueSize = size;

      if (!readActive && ingestQueueSize < MAX_QUEUE_SIZE$1) {
        // get the next response if there are more docs to fetch
        var sc = await scroll(scrollId); // eslint-disable-line no-await-in-loop,no-underscore-dangle,max-len
        scrollId = sc._scroll_id; // eslint-disable-line no-underscore-dangle
        responseQueue.push(sc);
        processResponseQueue();
      }
    });

    indexer.queueEmitter.on('resume', async function () {
      ingestQueueSize = 0;

      if (readActive) {
        return;
      }

      // get the next response if there are more docs to fetch
      var sc = await scroll(scrollId); // eslint-disable-line no-await-in-loop,no-underscore-dangle,max-len
      scrollId = sc._scroll_id; // eslint-disable-line no-underscore-dangle
      responseQueue.push(sc);
      processResponseQueue();
    });

    indexer.queueEmitter.on('finish', function () {
      progressBar.stop();
    });

    processResponseQueue();
  };
}

async function transformer(ref) {
  var deleteIndex = ref.deleteIndex; if ( deleteIndex === void 0 ) deleteIndex = false;
  var sourceClientConfig = ref.sourceClientConfig;
  var targetClientConfig = ref.targetClientConfig;
  var bufferSize = ref.bufferSize; if ( bufferSize === void 0 ) bufferSize = DEFAULT_BUFFER_SIZE;
  var fileName = ref.fileName;
  var splitRegex = ref.splitRegex; if ( splitRegex === void 0 ) splitRegex = /\n/;
  var sourceIndexName = ref.sourceIndexName;
  var targetIndexName = ref.targetIndexName;
  var mappings = ref.mappings;
  var mappingsOverride = ref.mappingsOverride; if ( mappingsOverride === void 0 ) mappingsOverride = false;
  var indexMappingTotalFieldsLimit = ref.indexMappingTotalFieldsLimit;
  var populatedFields = ref.populatedFields; if ( populatedFields === void 0 ) populatedFields = false;
  var query = ref.query;
  var skipHeader = ref.skipHeader; if ( skipHeader === void 0 ) skipHeader = false;
  var transform = ref.transform;
  var verbose = ref.verbose; if ( verbose === void 0 ) verbose = true;

  if (typeof targetIndexName === 'undefined') {
    throw Error('targetIndexName must be specified.');
  }

  var defaultClientConfig = {
    node: 'http://localhost:9200',
  };

  var sourceClient = new elasticsearch.Client(sourceClientConfig || defaultClientConfig);
  var targetClient = new elasticsearch.Client(
    targetClientConfig || sourceClientConfig || defaultClientConfig
  );

  var createMapping = createMappingFactory({
    sourceClient: sourceClient,
    sourceIndexName: sourceIndexName,
    targetClient: targetClient,
    targetIndexName: targetIndexName,
    mappings: mappings,
    mappingsOverride: mappingsOverride,
    indexMappingTotalFieldsLimit: indexMappingTotalFieldsLimit,
    verbose: verbose,
  });
  var indexer = indexQueueFactory({
    targetClient: targetClient,
    targetIndexName: targetIndexName,
    bufferSize: bufferSize,
    skipHeader: skipHeader,
    verbose: verbose,
  });

  function getReader() {
    if (typeof fileName !== 'undefined' && typeof sourceIndexName !== 'undefined') {
      throw Error('Only either one of fileName or sourceIndexName can be specified.');
    }

    if (typeof fileName === 'undefined' && typeof sourceIndexName === 'undefined') {
      throw Error('Either fileName or sourceIndexName must be specified.');
    }

    if (typeof fileName !== 'undefined') {
      return fileReaderFactory(indexer, fileName, transform, splitRegex, verbose);
    }

    if (typeof sourceIndexName !== 'undefined') {
      return indexReaderFactory(
        indexer,
        sourceIndexName,
        transform,
        sourceClient,
        query,
        bufferSize,
        populatedFields
      );
    }

    return null;
  }

  var reader = getReader();

  try {
    var indexExists = await targetClient.indices.exists({ index: targetIndexName });

    if (indexExists === false) {
      await createMapping();
      reader();
    } else if (deleteIndex === true) {
      await targetClient.indices.delete({ index: targetIndexName });
      await createMapping();
      reader();
    } else {
      reader();
    }
  } catch (error) {
    console.error('Error checking index existence:', error);
  } finally {
    // targetClient.close();
  }

  return { events: indexer.queueEmitter };
}

export default transformer;
